# https://lab.nexedi.com/nexedi/erp5/snippets/475
from collections import defaultdict
from lib2to3.fixer_base import BaseFix
import lib2to3.fixer_util
import lib2to3.pgen2
from lib2to3.pygram import python_symbols as syms
from lib2to3.pytree import Node

from my2to3.trace import create_table, register_tracing_function
from my2to3.util import parse_type


# id is used to differentiate the divisions of the same line.
insert_trace = create_table("division_trace", "filename", "lineno", "id", "dividend_type", "divisor_type")
insert_modified = create_table("division_modified", "filename", "lineno", "id")


@register_tracing_function
def division_traced(filename, lineno, id_, dividend, divisor):
  insert_trace(
    filename,
    lineno,
    id_,
    parse_type(type(dividend)),
    parse_type(type(divisor))
  )
  return dividend / divisor


class FixDivisionTrace(BaseFix):
  """Rewrites a / b into division_traced(id, a, b)
  """
  def start_tree(self, tree, filename):
    super(FixDivisionTrace, self).start_tree(tree, filename)
    self.ids = defaultdict(int)

  def match(self, node):
    if node.type == syms.term:
      # List the position of '/' nodes
      return [i for i, child in enumerate(node.children)
              if child.type == lib2to3.pgen2.token.SLASH]

  def transform(self, node, results):
    # Transforming "a / b" into "division_traced(id, a, b) is fairly trivial.
    # However, this is not the case when there are operations with the same
    # priority as the division (i.e. division, multiplication) which are
    # "in line", e.g.
    # - "a / b / c"
    # - "a * b / c * d / e"
    # - ...
    # That's why the logic is more complex here.

    previous_node = None

    for operator_pos in range(1, len(node.children), 2):

      children = [leaf for leaf in node.children[operator_pos - 1:operator_pos + 2]]
      if previous_node:
        # The first child is actually the previous operation
        children[0] = previous_node
      children = [leaf.clone() for leaf in children]

      if operator_pos in results:
        # It's a division operation
        lineno = node.get_lineno()
        dividend = children[0]
        if node.prefix == dividend.prefix:
          dividend.prefix = ''
        comma = lib2to3.fixer_util.Comma()
        comma.prefix = children[1].prefix

        id_ = self.ids[lineno]
        insert_modified(self.filename, lineno, id_)

        new_node = lib2to3.fixer_util.Call(
          lib2to3.fixer_util.Name("division_traced"),
          args=(
            lib2to3.fixer_util.String('"%s"' % self.filename),
            lib2to3.fixer_util.Comma(),
            lib2to3.fixer_util.Number(lineno),
            lib2to3.fixer_util.Comma(),
            lib2to3.fixer_util.Number(id_),
            lib2to3.fixer_util.Comma(),
            dividend,
            comma,
            children[2].clone(),
          ),
          prefix=node.prefix)

        self.ids[lineno] += 1

      else:
        # It's not a division operation
        new_node = Node(syms.term, children)

      previous_node = new_node

    node.replace(previous_node)
